/*
 * Copyright 1995-2008 Sun Microsystems, Inc.  All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

package java.lang;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.concurrent.Executors;
import java.util.concurrent.Executor;
import java.util.concurrent.ThreadFactory;
import java.util.StringTokenizer;
import java.security.AccessController;
import static java.security.AccessController.doPrivileged;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import sun.security.action.GetPropertyAction;

/* java.lang.Process subclass in the UNIX environment.
 *
 * @author Mario Wolczko and Ross Knippel.
 * @author Konstantin Kladko (ported to Linux)
 */

final class UNIXProcess extends Process {
    private static final sun.misc.JavaIOFileDescriptorAccess fdAccess
        = sun.misc.SharedSecrets.getJavaIOFileDescriptorAccess();

    private final int pid;
    private int exitcode;
    private boolean hasExited;

    private /* final */ OutputStream stdin;
    private /* final */ InputStream  stdout;
    private /* final */ InputStream  stderr;

    /* this is for the reaping thread */
    private native int waitForProcessExit(int pid);

    /**
     * Create a process using fork(2) and exec(2).
     *
     * @param std_fds an array of three file descriptors.
     *        Indexes 0, 1, and 2 correspond to standard input,
     *        standard output and standard error, respectively.  On
     *        input, a value of -1 means to create a pipe to connect
     *        child and parent processes.  On output, a value which
     *        is not -1 is the parent pipe fd corresponding to the
     *        pipe which has been created.  An element of this array
     *        is -1 on input if and only if it is <em>not</em> -1 on
     *        output.
     * @return the pid of the subprocess
     */
    private native int forkAndExec(byte[] prog,
                                   byte[] argBlock, int argc,
                                   byte[] envBlock, int envc,
                                   byte[] dir,
                                   int[] std_fds,
                                   boolean redirectErrorStream)
        throws IOException;

    /**
     * The thread factory used to create "process reaper" daemon threads.
     */
    private static class ProcessReaperThreadFactory implements ThreadFactory {
        private final static ThreadGroup group = getRootThreadGroup();
	 private static final String reaperName = "process reaper";

        private static ThreadGroup getRootThreadGroup() {
            return doPrivileged(new PrivilegedAction<ThreadGroup> () {
                public ThreadGroup run() {
                    ThreadGroup root = Thread.currentThread().getThreadGroup();
                    while (root.getParent() != null)
                        root = root.getParent();
                    return root;
                }});
        }

        public Thread newThread(Runnable grimReaper) {
            // Our thread stack requirement is quite modest.
            Thread t = new Thread(group, grimReaper, "process reaper", 32768);
            t.setDaemon(true);
            // A small attempt (probably futile) to avoid priority inversion
            t.setPriority(Thread.MAX_PRIORITY);
            return t;
        }
    }

    /**
     * The thread pool of "process reaper" daemon threads.
     */
    private static final Executor processReaperExecutor =
        doPrivileged(new PrivilegedAction<Executor>() {
            public Executor run() {
                return Executors.newCachedThreadPool
                    (new ProcessReaperThreadFactory());
            }});

    UNIXProcess(final byte[] prog,
                final byte[] argBlock, final int argc,
                final byte[] envBlock, final int envc,
                final byte[] dir,
                final int[] std_fds,
                final boolean redirectErrorStream)
        throws IOException {

        /*ibm@JSE-8409 start*//*IBM-ios start*/
        if(os400 == 1) {
        		       
            String filename = new String(prog,platformEncoding);
            final byte fullprog[] = fullPath(prog);
            filename = new String(fullprog,platformEncoding);
            pid = forkAndExec(fullprog,
                              argBlock, argc,
                              envBlock, envc,
                              dir,
                              std_fds,
                              redirectErrorStream);
        }
        /*ibm@JSE-8409 end*//*IBM-ios end*/
        else {
            pid = forkAndExec(prog,
                              argBlock, argc,
                              envBlock, envc,
                              dir,
                              std_fds,
                              redirectErrorStream);
        }
                          	
        try {
            doPrivileged(new PrivilegedExceptionAction<Void>() {
                public Void run() throws IOException {
                    initStreams(std_fds);
                    return null;
                }});
        } catch (PrivilegedActionException ex) {
            throw (IOException) ex.getException();
        }
    }

    static FileDescriptor newFileDescriptor(int fd) {
        FileDescriptor fileDescriptor = new FileDescriptor();
        fdAccess.set(fileDescriptor, fd);
        return fileDescriptor;
    }

    void initStreams(int[] fds) throws IOException {
        stdin = (fds[0] == -1) ?
            ProcessBuilder.NullOutputStream.INSTANCE :
            new BufferedOutputStream(new FileOutputStream(newFileDescriptor(fds[0])));
	
        stdout = (fds[1] == -1) ?
            ProcessBuilder.NullInputStream.INSTANCE :
			new BufferedInputStream(new DeferredCloseInputStream(newFileDescriptor(fds[1])));

        stderr = (fds[2] == -1) ?
            ProcessBuilder.NullInputStream.INSTANCE :
            new DeferredCloseInputStream(newFileDescriptor(fds[2]));

        processReaperExecutor.execute(new Runnable() {
            public void run() {
                int exitcode = waitForProcessExit(pid);
                UNIXProcess.this.processExited(exitcode);
            }});
    }

    synchronized void processExited(int exitcode) {
        this.exitcode = exitcode;
        hasExited = true;
        notifyAll();
    }

    public OutputStream getOutputStream() {
        return stdin;
    }

    public InputStream getInputStream() {
        return stdout;
    }

    public InputStream getErrorStream() {
        return stderr;
    }

    public synchronized int waitFor() throws InterruptedException {
        while (!hasExited) {
            wait();
        }
        return exitcode;
    }

    public synchronized int exitValue() {
        if (!hasExited) {
            throw new IllegalThreadStateException("process hasn't exited");
        }
        return exitcode;
    }

    private static native void destroyProcess(int pid);
    public void destroy() {
        // There is a risk that pid will be recycled, causing us to
        // kill the wrong process!  So we only terminate processes
        // that appear to still be running.  Even with this check,
        // there is an unavoidable race condition here, but the window
        // is very small, and OSes try hard to not recycle pids too
        // soon, so this is quite safe.
        synchronized (this) {
            if (!hasExited)
                destroyProcess(pid);
        }
        try { stdin.close();  } catch (IOException ignored) {}
        try { stdout.close(); } catch (IOException ignored) {}
        try { stderr.close(); } catch (IOException ignored) {}
    }

/*ibm@JSE-8409 start*//*IBM-ios start*/
    private native int statExecutable(byte []filename_bytes);
    	
    /**
     * Returns the full path name for the specified filename.
     */
    public byte[] fullPath(byte []filename_bytes) throws IOException {
    	
        String sep = (String)AccessController.doPrivileged           /*ibm@87093*/
                         (new GetPropertyAction("file.separator"));  /*ibm@87093*/
        String PATH = ProcessEnvironment.getenv("PATH");
        String ps = (String)AccessController.doPrivileged            /*ibm@87093*/
                         (new GetPropertyAction("path.separator"));  /*ibm@87093*/
        String fullfilename = null;
        byte fullfilename_bytes[] = null;

        String filename;
        /* unterminatedFilename contains filename without the null character */
        String unterminatedFilename;                             /*ibm@133117*/

        try {
            filename = new String(filename_bytes,platformEncoding);
        } catch (UnsupportedEncodingException e) {
            filename = new String(filename_bytes);
        }
        unterminatedFilename = filename.replace("\u0000","");    /*ibm@133117*/


        /* If the filename we want to exec has any slashes in it then  */
        /* we shouldn't do a path search, as in /foo ./foo or foo/bar. */
        if ((filename.indexOf(sep) < 0) && PATH != null) {
        /*ibm@138341 start*/
        if (os400 == 1) {
        /* Prepend java.home to the PATH for command resolution. i5/OS needs */
        /* to check the java.home location before anything in the PATH.      */
        String java_home = (String)AccessController.doPrivileged
                          (new GetPropertyAction("java.home"));
        if (java_home != null) {
          StringBuffer PREPEND = new StringBuffer();
          PREPEND.append(java_home + sep + ".." + sep + "bin" + ps);
          PREPEND.append(java_home + sep + "bin" + ps);
          PATH = PREPEND.toString() + PATH;
          if (ProcessEnvironment.getenv("QIBM_QYJ9NTVUTL_DEBUG") != null) {
              System.out.println("UNIXProcess.java input pprog=" + filename);
              System.out.println("UNIXProcess.java PATH checked=" + PATH);
           }
        }
      }
      /*ibm@138341 end*/

       StringTokenizer st = new StringTokenizer(PATH,ps);
       boolean foundit = false;
       while (st.hasMoreTokens() && !foundit) {
           fullfilename = st.nextToken() + sep + filename;
           try {
               fullfilename_bytes =
                   fullfilename.getBytes(platformEncoding);
           } catch (UnsupportedEncodingException e) {
               fullfilename_bytes = fullfilename.getBytes();
           }
           if ( statExecutable(fullfilename_bytes) == 0) {
               foundit = true;
           }
     }
       if (!foundit) throw new IOException(unterminatedFilename +
                         ": not found");       /*ibm@133117*/

     } else {
            int status = statExecutable(filename_bytes);
            switch (status) {
                case -1: throw new IOException(unterminatedFilename +
                                         ": not found");        /*ibm@133117*/
                case -2: throw new IOException(unterminatedFilename+
                                         ": cannot execute");   /*ibm@133117*/
            }
            fullfilename_bytes = filename_bytes;
        }

        return fullfilename_bytes;
    }
   

/*ibm@JSE-8409 end*//*IBM-ios end*/
    // A FileInputStream that supports the deferment of the actual close
    // operation until the last pending I/O operation on the stream has
    // finished.  This is required on Solaris because we must close the stdin
    // and stdout streams in the destroy method in order to reclaim the
    // underlying file descriptors.  Doing so, however, causes any thread
    // currently blocked in a read on one of those streams to receive an
    // IOException("Bad file number"), which is incompatible with historical
    // behavior.  By deferring the close we allow any pending reads to see -1
    // (EOF) as they did before.
    //
    private static class DeferredCloseInputStream
        extends FileInputStream
    {

        private DeferredCloseInputStream(FileDescriptor fd) {
            super(fd);
        }

        private class InterruptLock {}
        private Object lock = new InterruptLock();      // For the following fields
        private boolean closePending = false;
        private int useCount = 0;
        private InputStream streamToClose;

        private void raise() {
            synchronized (lock) {
                useCount++;
            }
        }

        private void lower() throws IOException {
            synchronized (lock) {
                useCount--;
                if (useCount == 0 && closePending) {
                    streamToClose.close();
                }
            }
        }

        // stc is the actual stream to be closed; it might be this object, or
        // it might be an upstream object for which this object is downstream.
        //
        private void closeDeferred(InputStream stc) throws IOException {
            synchronized (lock) {
                if (useCount == 0) {
                    stc.close();
                } else {
                    closePending = true;
                    streamToClose = stc;
                }
            }
        }

        public void close() throws IOException {
            synchronized (lock) {
                useCount = 0;
                closePending = false;
            }
            super.close();
        }

        public int read() throws IOException {
            raise();
            try {
                return super.read();
            } finally {
                lower();
            }
        }

        public int read(byte[] b) throws IOException {
            raise();
            try {
                return super.read(b);
            } finally {
                lower();
            }
        }

        public int read(byte[] b, int off, int len) throws IOException {
            raise();
            try {
                return super.read(b, off, len);
            } finally {
                lower();
            }
        }

        public long skip(long n) throws IOException {
            raise();
            try {
                return super.skip(n);
            } finally {
                lower();
            }
        }

        public int available() throws IOException {
            raise();
            try {
                return super.available();
            } finally {
                lower();
            }
        }

    }
	
    /* This routine initializes JNI field offsets for the class */
    private static native void initIDs();
    
    /*ibm@JSE-8409 start*//*IBM-ios start*/
    static String platformEncoding;
    private static int os400 = -1;    
    private static void initiOsFlag() {
       if (os400 < 0) {
		    String os_name = (String)AccessController.doPrivileged
		                          (new GetPropertyAction("os.name"));
		    os400 = ((os_name != null) && os_name.equals("OS/400")) ? 1 : 0;
		   
        platformEncoding = (String)AccessController.doPrivileged
                               (new GetPropertyAction("ibm.system.encoding"));
        if (platformEncoding == null) {
            platformEncoding = (String)AccessController.doPrivileged
                                   (new GetPropertyAction("console.encoding"));
        }
        if (platformEncoding == null) {
            platformEncoding = (String)AccessController.doPrivileged
                                   (new GetPropertyAction("file.encoding"));
                                  }
	    }
    }
    /*ibm@JSE-8409 end*//*IBM-ios end*/
    static {
        initIDs();
        initiOsFlag();        /*ibm@JSE-8409 *//*IBM-ios_bringup*/
    }
}
